# coding: utf-8

from typing import List
from copy import deepcopy
from supervisely.imaging.color import random_rgb, rgb2hex, hex2rgb, _validate_color
from supervisely.io.json import JsonSerializable
from supervisely.collection.key_indexed_collection import KeyObject
from supervisely._utils import take_with_default


class TagValueType:
    NONE = 'none'
    ANY_NUMBER = 'any_number'
    ANY_STRING = 'any_string'
    ONEOF_STRING = 'oneof_string'


class TagMetaJsonFields:
    ID = 'id'
    NAME = 'name'
    VALUE_TYPE = 'value_type'
    VALUES = 'values'
    COLOR = 'color'
    APPLICABLE_TYPE = 'applicable_type'
    HOTKEY = "hotkey"
    APPLICABLE_CLASSES = 'classes'


class TagApplicableTo:
    ALL = 'all' # both images and objects
    IMAGES_ONLY = 'imagesOnly'
    OBJECTS_ONLY = 'objectsOnly'


SUPPORTED_TAG_VALUE_TYPES = [TagValueType.NONE, TagValueType.ANY_NUMBER, TagValueType.ANY_STRING, TagValueType.ONEOF_STRING]
SUPPORTED_APPLICABLE_TO = [TagApplicableTo.ALL, TagApplicableTo.IMAGES_ONLY, TagApplicableTo.OBJECTS_ONLY]


class TagMeta(KeyObject, JsonSerializable):
    '''
    This is a class for creating and using TagMeta objects. It include tag name, value type, and possible values for
    tags with enum values.
    '''
    def __init__(self, name: str, value_type: str, possible_values: List[str] = None, color: List[int]=None, sly_id=None,
                 hotkey: str = None, applicable_to: str = None, applicable_classes: List[str]=None):
        """
        :param name: str
        :param value_type: str (one of TagValueType fields)
        :param values: list of possible values (i.e. [str]) or None
        :param color: [R, G, B]
        """

        if value_type not in SUPPORTED_TAG_VALUE_TYPES:
            raise ValueError("value_type = {!r} is unknown, should be one of {}"
                             .format(value_type, SUPPORTED_TAG_VALUE_TYPES))

        self._name = name
        self._value_type = value_type
        self._possible_values = possible_values
        self._color = random_rgb() if color is None else deepcopy(color)
        self._sly_id = sly_id
        self._hotkey = take_with_default(hotkey, "")
        self._applicable_to = take_with_default(applicable_to, TagApplicableTo.ALL)
        self._applicable_classes = take_with_default(applicable_classes, [])
        if self._applicable_to not in SUPPORTED_APPLICABLE_TO:
            raise ValueError("applicable_to = {!r} is unknown, should be one of {}"
                             .format(self._applicable_to, SUPPORTED_APPLICABLE_TO))

        if self._value_type == TagValueType.ONEOF_STRING:
            if self._possible_values is None:
                raise ValueError("TagValueType is ONEOF_STRING. List of possible values have to be defined.")
            if not all(isinstance(item, str) for item in self._possible_values):
                raise ValueError("TagValueType is ONEOF_STRING. All possible values have to be strings")
        elif self._possible_values is not None:
            raise ValueError("TagValueType is {!r}. possible_values variable have to be None".format(self._value_type))

        _validate_color(self._color)

    @property
    def name(self):
        return self._name

    def key(self):
        return self.name

    @property
    def value_type(self):
        return self._value_type

    @property
    def possible_values(self):
        return self._possible_values.copy() if self._possible_values is not None else None

    @property
    def color(self):
        return self._color.copy()

    @property
    def sly_id(self):
        return self._sly_id

    @property
    def hotkey(self):
        return self._hotkey

    @property
    def applicable_to(self):
        return self._applicable_to

    @property
    def applicable_classes(self):
        return self._applicable_classes

    def to_json(self):
        '''
        The function to_json convert TagMeta object to json format
        :return: tagmeta in json format
        '''
        jdict = {
            TagMetaJsonFields.NAME: self.name,
            TagMetaJsonFields.VALUE_TYPE: self.value_type,
            TagMetaJsonFields.COLOR: rgb2hex(self.color)
        }
        if self.value_type == TagValueType.ONEOF_STRING:
            jdict[TagMetaJsonFields.VALUES] = self.possible_values

        if self.sly_id is not None:
            jdict[TagMetaJsonFields.ID] = self.sly_id
        if self._hotkey is not None:
            jdict[TagMetaJsonFields.HOTKEY] = self.hotkey
        if self._applicable_to is not None:
            jdict[TagMetaJsonFields.APPLICABLE_TYPE] = self.applicable_to
        if self._applicable_classes is not None:
            jdict[TagMetaJsonFields.APPLICABLE_CLASSES] = self.applicable_classes

        return jdict

    @classmethod
    def from_json(cls, data):
        '''
        The function from_json convert tagmeta from json format to TagMeta class object.
        :param data: input tagmeta in json format
        :return: TagMeta class object
        '''
        if isinstance(data, str):
            return cls(name=data, value_type=TagValueType.NONE)
        elif isinstance(data, dict):
            name = data[TagMetaJsonFields.NAME]
            value_type = data[TagMetaJsonFields.VALUE_TYPE]
            values = data.get(TagMetaJsonFields.VALUES)
            color = data.get(TagMetaJsonFields.COLOR)
            if color is not None:
                color = hex2rgb(color)
            sly_id = data.get(TagMetaJsonFields.ID, None)

            hotkey = data.get(TagMetaJsonFields.HOTKEY, "")
            applicable_to = data.get(TagMetaJsonFields.APPLICABLE_TYPE, TagApplicableTo.ALL)
            applicable_classes = data.get(TagMetaJsonFields.APPLICABLE_CLASSES, [])

            return cls(name=name, value_type=value_type, possible_values=values, color=color, sly_id=sly_id,
                       hotkey=hotkey, applicable_to=applicable_to, applicable_classes=applicable_classes)
        else:
            raise ValueError('Tags must be dict or str types.')

    def add_possible_value(self, value):
        '''
        The function add_possible_value add new value to the list of possible_values. If value_type is not 'oneof_string'
        it generate ValueError error.
        :param value: value added
        :return: TagMeta class object
        '''
        if self.value_type is TagValueType.ONEOF_STRING:
            if value in self._possible_values:
                raise ValueError('Value {} already exists for tag {}'.format(value, self.name))
            else:
                return self.clone(possible_values=[*self.possible_values, value])
        else:
            raise ValueError("Tag {!r} has type {!r}. Possible value can be added only to oneof_string".format(self.name, self.value_type))

    def is_valid_value(self, value):
        '''
        The function is_valid_value cross-checked value against the value_type to make sure the value is valid.
        If value is unsupported it generate ValueError error.
        :return: True if value is supported, False in other way.
        '''
        if self.value_type == TagValueType.NONE:
            return value is None
        elif self.value_type == TagValueType.ANY_NUMBER:
            return isinstance(value, (int, float))
        elif self.value_type == TagValueType.ANY_STRING:
            return isinstance(value, str)
        elif self.value_type == TagValueType.ONEOF_STRING:
            return isinstance(value, str) and (value in self._possible_values)
        else:
            raise ValueError('Unsupported TagValueType detected ({})!'.format(self.value_type))

    def __eq__(self, other):
        # TODO compare colors also here (need to check the usages and replace with is_compatible() where appropriate).
        return (isinstance(other, TagMeta) and
                self.name == other.name and
                self.value_type == other.value_type and
                self.possible_values == other.possible_values)

    def __ne__(self, other):
        return not self == other

    def is_compatible(self, other):
        '''
        The function is_compatible checks the input data against the given TagMeta object
        :param other: some Object
        '''
        return (isinstance(other, TagMeta) and
                self.name == other.name and
                self.value_type == other.value_type and
                self.possible_values == other.possible_values)

    def clone(self, name=None, value_type=None, possible_values=None, color=None, sly_id=None,
              hotkey=None, applicable_to=None, applicable_classes=None):
        '''
        The function clone make copy of the TagMeta class object
        :return: TagMeta class object
        '''
        return TagMeta(name=take_with_default(name, self.name),
                       value_type=take_with_default(value_type, self.value_type),
                       possible_values=take_with_default(possible_values, self.possible_values),
                       color=take_with_default(color, self.color),
                       sly_id=take_with_default(sly_id, self.sly_id),
                       hotkey=take_with_default(hotkey, self.hotkey),
                       applicable_to=take_with_default(applicable_to, self.applicable_to),
                       applicable_classes=take_with_default(applicable_classes, self.applicable_classes))

    def __str__(self):
        return "{:<7s}{:<24} {:<7s}{:<13} {:<13s}{:<10} {:<13s}{:<10} {:<13s}{:<10} {:<13s}{}".format(
            'Name:', self.name, 'Value type:', self.value_type, 'Possible values:', str(self.possible_values),
            'Hotkey', self.hotkey, 'Applicable to', self.applicable_to, 'Applicable classes', self.applicable_classes)

    @classmethod
    def get_header_ptable(cls):
        return ['Name', 'Value type', 'Possible values', 'Hotkey', 'Applicable to', 'Applicable classes']

    def get_row_ptable(self):
        '''
        :return: information about TagMeta class object(name of meta, type value, and list of possible values)
        '''
        return [self.name, self.value_type, self.possible_values, self.hotkey, self.applicable_to, self.applicable_classes]


# coding: utf-8
import os
import shutil
import tarfile
from pathlib import Path
import urllib
from supervisely.api.module_api import ModuleApiBase, ApiField
from supervisely.io.fs import ensure_base_path, get_file_name_with_ext
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
import mimetypes
from supervisely.io.fs import get_file_ext, get_file_name, list_files_recursively
from supervisely.imaging.image import write_bytes, get_hash
from supervisely.task.progress import Progress
from supervisely.io.fs_cache import FileCache
from supervisely.io.fs import get_file_hash, get_file_ext, get_file_size, list_files_recursively, silent_remove

class FileApi(ModuleApiBase):
    @staticmethod
    def info_sequence():
        return [
            ApiField.TEAM_ID,
            ApiField.ID,
            ApiField.USER_ID,
            ApiField.NAME,
            ApiField.HASH,
            ApiField.PATH,
            ApiField.STORAGE_PATH,
            ApiField.MIME2,
            ApiField.EXT2,
            ApiField.SIZEB2,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
            ApiField.FULL_STORAGE_URL,
        ]

    @staticmethod
    def info_tuple_name():
        return 'FileInfo'

    def list(self, team_id, path):
        response = self._api.post('file-storage.list', {ApiField.TEAM_ID: team_id, ApiField.PATH: path})
        return response.json()

    def list2(self, team_id, path):
        response = self._api.post('file-storage.list', {ApiField.TEAM_ID: team_id, ApiField.PATH: path})
        results = [self._convert_json_info(info_json) for info_json in response.json()]
        return results

    def get_directory_size(self, team_id, path):
        dir_size = 0
        file_infos = self.list2(team_id, path)
        for file_info in file_infos:
            dir_size += file_info.sizeb
        return dir_size

    def _download(self, team_id, remote_path, local_save_path, progress_cb=None):  # TODO: progress bar
        response = self._api.post('file-storage.download', {ApiField.TEAM_ID: team_id, ApiField.PATH: remote_path}, stream=True)
        #print(response.headers)
        #print(response.headers['Content-Length'])
        ensure_base_path(local_save_path)
        with open(local_save_path, 'wb') as fd:
            for chunk in response.iter_content(chunk_size=1024 * 1024):
                fd.write(chunk)
                if progress_cb is not None:
                    progress_cb(len(chunk))

    def download(self, team_id, remote_path, local_save_path, cache: FileCache = None, progress_cb=None):
        if cache is None:
            self._download(team_id, remote_path, local_save_path, progress_cb)
        else:
            file_info = self.get_info_by_path(team_id, remote_path)
            if file_info.hash is None:
                self._download(team_id, remote_path, local_save_path, progress_cb)
            else:
                cache_path = cache.check_storage_object(file_info.hash, get_file_ext(remote_path))
                if cache_path is None:
                    # file not in cache
                    self._download(team_id, remote_path, local_save_path, progress_cb)
                    if file_info.hash != get_file_hash(local_save_path):
                        raise KeyError(f"Remote and local hashes are different (team id: {team_id}, file: {remote_path})")
                    cache.write_object(local_save_path, file_info.hash)
                else:
                    cache.read_object(file_info.hash, local_save_path)
                    if progress_cb is not None:
                        progress_cb(get_file_size(local_save_path))

    def download_directory(self, team_id, remote_path, local_save_path, progress_cb=None):
        local_temp_archive = os.path.join(local_save_path, "temp.tar")
        self._download(team_id, remote_path, local_temp_archive, progress_cb)
        tr = tarfile.open(local_temp_archive)
        tr.extractall(local_save_path)
        silent_remove(local_temp_archive)
        temp_dir = os.path.join(local_save_path, os.path.basename(os.path.normpath(remote_path)))
        file_names = os.listdir(temp_dir)
        for file_name in file_names:
            shutil.move(os.path.join(temp_dir, file_name), local_save_path)
        shutil.rmtree(temp_dir)


    def _upload_legacy(self, team_id, src, dst):
        def path_to_bytes_stream(path):
            return open(path, 'rb')
        item = get_file_name_with_ext(dst)
        content_dict = {}
        content_dict[ApiField.NAME] = item

        dst_dir = os.path.dirname(dst)
        if not dst_dir.endswith(os.path.sep):
            dst_dir += os.path.sep
        content_dict[ApiField.PATH] = dst_dir # os.path.basedir ...
        content_dict["file"] = (item, path_to_bytes_stream(src), mimetypes.MimeTypes().guess_type(src)[0])
        encoder = MultipartEncoder(fields=content_dict)
        resp = self._api.post("file-storage.upload?teamId={}".format(team_id), encoder)
        return resp.json()

    def upload(self, team_id, src, dst, progress_cb=None):
        return self.upload_bulk(team_id, [src], [dst], progress_cb)[0]

    def upload_bulk(self, team_id, src_paths, dst_paths, progress_cb=None):
        def path_to_bytes_stream(path):
            return open(path, 'rb')
        content_dict = []
        for idx, (src, dst) in enumerate(zip(src_paths, dst_paths)):
            name = get_file_name_with_ext(dst)
            content_dict.append((ApiField.NAME, name))
            dst_dir = os.path.dirname(dst)
            if not dst_dir.endswith(os.path.sep):
                dst_dir += os.path.sep
            content_dict.append((ApiField.PATH, dst_dir))
            content_dict.append(("file", (name, path_to_bytes_stream(src), mimetypes.MimeTypes().guess_type(src)[0])))
        encoder = MultipartEncoder(fields=content_dict)

        # progress = None
        # if progress_logger is not None:
        #     progress = Progress("Uploading", encoder.len, progress_logger, is_size=True)
        #
        # def _print_progress(monitor):
        #     if progress is not None:
        #         progress.set_current_value(monitor.bytes_read)
        #         print(monitor.bytes_read, monitor.len)
        # last_percent = 0
        # def _progress_callback(monitor):
        #     cur_percent = int(monitor.bytes_read * 100.0 / monitor.len)
        #     if cur_percent - last_percent > 10 or cur_percent == 100:
        #         api.task.set_fields(task_id, [{"field": "data.previewProgress", "payload": cur_percent}])
        #     last_percent = cur_percent

        if progress_cb is None:
            data = encoder
        else:
            data = MultipartEncoderMonitor(encoder, progress_cb)
        resp = self._api.post("file-storage.bulk.upload?teamId={}".format(team_id), data)
        results = [self._convert_json_info(info_json) for info_json in resp.json()]
        return results

    def rename(self, old_name, new_name):
        pass

    def remove(self, team_id, path):
        resp = self._api.post("file-storage.remove",{ApiField.TEAM_ID: team_id, ApiField.PATH: path})

    def exists(self, team_id, remote_path):
        path_infos = self.list(team_id, remote_path)
        for info in path_infos:
            if info["path"] == remote_path:
                return True
        return False

    def dir_exists(self, team_id, remote_directory):
        files_infos = self.list(team_id, remote_directory)
        if len(files_infos) > 0:
            return True
        return False

    def get_free_name(self, team_id, path):
        directory = Path(path).parent
        name = get_file_name(path)
        ext = get_file_ext(path)
        res_name = name
        suffix = 0

        def _combine(suffix:int=None):
            res = "{}/{}".format(directory, res_name)
            if suffix is not None:
                res += "_{:03d}".format(suffix)
            if ext:
                res += "{}".format(ext)
            return res

        res_path = _combine()
        while self.exists(team_id, res_path):
            res_path = _combine(suffix)
            suffix += 1
        return res_path

    def get_url(self, file_id):
        return urllib.parse.urljoin(self._api.server_address, "files/{}".format(file_id))

    def get_info_by_path(self, team_id, remote_path):
        path_infos = self.list(team_id, remote_path)
        for info in path_infos:
            if info["path"] == remote_path:
                return self._convert_json_info(info)
        return None

    def _convert_json_info(self, info: dict, skip_missing=True):
        res = super()._convert_json_info(info, skip_missing=skip_missing)
        #if res.storage_path is not None:
        #    res = res._replace(full_storage_url=urllib.parse.urljoin(self._api.server_address, res.storage_path))
        return res

    def get_info_by_id(self, id):
        resp = self._api.post('file-storage.info', {ApiField.ID: id})
        return self._convert_json_info(resp.json())

    def get_free_dir_name(self, team_id, dir_path):
        res_dir = dir_path.rstrip('/')
        suffix = 1
        while self.dir_exists(team_id, res_dir):
            res_dir = dir_path.rstrip('/') + f"_{suffix:03d}"
            suffix += 1
        return res_dir

    def upload_directory(self, team_id, local_dir, remote_dir,
                         change_name_if_conflict=True, progress_size_cb=None):
        if self.dir_exists(team_id, remote_dir):
            if change_name_if_conflict is True:
                res_remote_dir = self.get_free_dir_name(team_id, remote_dir)
            else:
                raise FileExistsError(f"Directory {remote_dir} already exists in your team (id={team_id})")
        else:
            res_remote_dir = remote_dir

        local_files = list_files_recursively(local_dir)
        remote_files = [file.replace(local_dir, res_remote_dir) for file in local_files]

        upload_results = self.upload_bulk(team_id, local_files, remote_files, progress_size_cb)
        return res_remote_dir




from pathlib import Path

from varname import varname
from jinja2 import Environment
import markupsafe
from supervisely.app.jinja2 import create_env
from supervisely.app.content import DataJson, StateJson
from supervisely.app.fastapi import Jinja2Templates


class Widget:
    def __init__(self, widget_id: str = None, file_path: str = __file__):
        self.widget_id = widget_id
        self._file_path = file_path
        if self.widget_id is None:
            self.widget_id = varname(frame=2)
        self._register()

    def _register(self):
        # get singletons
        data = DataJson()
        data.raise_for_key(self.widget_id)
        self.update_data()

        state = StateJson()
        state.raise_for_key(self.widget_id)
        self.update_state(state=state)

        templates = Jinja2Templates()
        templates.context_widgets[self.widget_id] = self

    def get_json_data(self):
        raise NotImplementedError()

    def get_json_state(self):
        raise NotImplementedError()

    def update_state(self, state):
        serialized_state = self.get_json_state()
        if serialized_state is not None:
            state[self.widget_id] = serialized_state

    def update_data(self):
        data = DataJson()
        serialized_data = self.get_json_data()
        if serialized_data is not None:
            data[self.widget_id] = serialized_data

    def to_html(self):
        current_dir = Path(self._file_path).parent.absolute()
        jinja2_sly_env: Environment = create_env(current_dir)
        html = jinja2_sly_env.get_template("template.html").render({"widget": self})
        return markupsafe.Markup(html)


# https://stackoverflow.com/questions/18425225/getting-the-name-of-a-variable-as-a-string
# https://github.com/pwwang/python-varname
# https://stackoverflow.com/questions/18425225/getting-the-name-of-a-variable-as-a-string/18425523#18425523
# https://ideone.com/ym3bkD
# https://github.com/pwwang/python-varname
# https://stackoverflow.com/questions/13034496/using-global-variables-between-files
# https://docs.python.org/3/faq/programming.html#how-do-i-share-global-variables-across-modules



# coding: utf-8

from supervisely.annotation.tag_meta import TagValueType
from supervisely.collection.key_indexed_collection import KeyObject
from supervisely._utils import take_with_default


class TagJsonFields:
    TAG_NAME = 'name'
    VALUE = 'value'
    LABELER_LOGIN = 'labelerLogin'
    UPDATED_AT = 'updatedAt'
    CREATED_AT = 'createdAt'
    ID = 'id'
    #TAG_META_ID = 'tagId'


class Tag(KeyObject):
    '''
    This is a class for creating and using Tags objects. The tags can be attached both to whole images and to
    individual geometric labels.
    '''
    def __init__(self, meta, value=None, sly_id=None, labeler_login=None, updated_at=None, created_at=None):
        '''
        The constructor for Tag class.
        :param meta: Tag metadata: it include tag name, value type, and possible values for tags with enum values.
        When creating a new tag, the value is automatically cross-checked against the metadata to make sure the value
        is valid.
        :param value: There are 3 possible value types of value: ANY_NUMBER for numeric values,
        ANY_STRING for arbitrary string values, ONEOF_STRING for string values restricted to a given whitelist/
        '''

        if meta is None:
            raise ValueError('TagMeta is None')
        self._meta = meta
        self._value = value
        if not self._meta.is_valid_value(value):
            raise ValueError('Tag {} can not have value {}'.format(self.meta.name, value))
        self.labeler_login = labeler_login
        self.updated_at = updated_at
        self.created_at = created_at
        self.sly_id = sly_id

    @property
    def meta(self):
        return self._meta

    @property
    def value(self):
        return self._value

    @property
    def name(self):
        return self._meta.name

    def key(self):
        return self._meta.key()

    def to_json(self):
        '''
        The function to_json convert tag to json format
        :return: tag in json format
        '''
        res = {
            TagJsonFields.TAG_NAME: self.meta.name,
            #TagJsonFields.VALUE: self.value
        }
        if self.meta.value_type != TagValueType.NONE:
            res[TagJsonFields.VALUE] = self.value
        if self.labeler_login is not None:
            res[TagJsonFields.LABELER_LOGIN] = self.labeler_login
        if self.updated_at is not None:
            res[TagJsonFields.UPDATED_AT] = self.updated_at
        if self.created_at is not None:
            res[TagJsonFields.CREATED_AT] = self.created_at
        # if self.meta.value_type == TagValueType.NONE:
        #     return self.meta.name
        # else:
        #     return {
        #         TagJsonFields.TAG_NAME: self.meta.name,
        #         TagJsonFields.VALUE: self.value
        #     }
        return res

    @classmethod
    def from_json(cls, data, tag_meta_collection):
        '''
        The function from_json convert tag from json format to Tag class object.
        :param data: input tag in json format
        :param tag_meta_collection: TagCollection class object
        :return: Tag class object
        '''
        if type(data) is str:
            tag_name = data
            value = None
            labeler_login = None
            updated_at = None
            created_at = None
            sly_id = None
        else:
            tag_name = data[TagJsonFields.TAG_NAME]
            value = data.get(TagJsonFields.VALUE, None)
            labeler_login = data.get(TagJsonFields.LABELER_LOGIN, None)
            updated_at = data.get(TagJsonFields.UPDATED_AT, None)
            created_at = data.get(TagJsonFields.CREATED_AT, None)
            sly_id = data.get(TagJsonFields.ID, None)
        meta = tag_meta_collection.get(tag_name)
        return cls(meta=meta, value=value, sly_id=sly_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)

    def get_compact_str(self):
        '''
        The function get_compact_str displays information about tag(name, value) in string format
        '''
        if (self.meta.value_type != TagValueType.NONE) and (len(str(self.value)) > 0):
            return '{}:{}'.format(self.name, self.value)
        return self.name

    def __eq__(self, other):
        return isinstance(other, Tag) and self.meta == other.meta and self.value == other.value

    def __ne__(self, other):
        return not self == other

    def clone(self, meta=None, value=None, sly_id=None, labeler_login=None, updated_at=None, created_at=None):
        '''
        The function clone make copy of the Tag class object
        :return: Tag class object
        '''
        return Tag(meta=take_with_default(meta, self.meta),
                   value=take_with_default(value, self.value),
                   sly_id=take_with_default(sly_id, self.sly_id),
                   labeler_login=take_with_default(labeler_login, self.labeler_login),
                   updated_at=take_with_default(updated_at, self.updated_at),
                   created_at=take_with_default(created_at, self.created_at))

    def __str__(self):
        return '{:<7s}{:<10}{:<7s} {:<13}{:<7s} {:<10}'.format('Name:', self._meta.name,
                                                               'Value type:', self._meta.value_type,
                                                               'Value:', str(self.value))

    @classmethod
    def get_header_ptable(cls):
        return ['Name', 'Value type', 'Value']

    def get_row_ptable(self):
        '''
        :return: information about Tag class object(name of meta, value of the tag, and value type)
        '''
        return [self._meta.name, self._meta.value_type, self.value]


# coding: utf-8

# fields names
ID = 'id'
KEY = 'key'
INDEX = 'index'
FIGURES = 'figures'
OBJECT_ID = 'objectId'
OBJECT_KEY = 'objectKey'
FRAME_RANGE = 'frameRange'

TAGS_MAP = 'tags'
OBJECTS_MAP = 'objects'
FIGURES_MAP = 'figures'
VIDEOS_MAP = 'videos'

FRAMES = 'frames'

OBJECTS = 'objects'
FRAMES_COUNT = 'framesCount'
VIDEO_ID = 'videoId'
VIDEO_NAME = 'videoName'
TAGS = 'tags'
DESCRIPTION = 'description'

IMG_DESCRIPTION = 'description'
IMG_SIZE = 'size'
IMG_SIZE_WIDTH = 'width'
IMG_SIZE_HEIGHT = 'height'

META = 'meta'



# coding: utf-8

from supervisely.api.module_api import ApiField
from supervisely.api.entity_annotation.tag_api import TagApi


class VideoTagApi(TagApi):
    _entity_id_field = ApiField.VIDEO_ID
    _method_bulk_add = 'videos.tags.bulk.add'

    def remove_from_video(self, tag_id):
        self._api.post('videos.tags.remove', {ApiField.ID: tag_id})

    def update_frame_range(self, tag_id, frame_range):
        self._api.post('videos.tags.update', {ApiField.ID: tag_id, ApiField.FRAME_RANGE: frame_range})

    def update_value(self, tag_id, tag_value):
        self._api.post('videos.tags.update-value', {ApiField.ID: tag_id, ApiField.VALUE: tag_value})

    def add_tag(self, project_meta_tag_id, video_id, value=None, frame_range=None):
        request_data = {
            ApiField.TAG_ID: project_meta_tag_id,
            ApiField.VIDEO_ID: video_id
        }
        if value:
            request_data[ApiField.VALUE] = value
        if frame_range:
            request_data[ApiField.FRAME_RANGE] = frame_range

        self._api.post('videos.tags.add', request_data)


"""
    import supervisely as sly

    app = sly.AppService()
    api = app.public_api

    video_info = api.video.get_info_by_id(1114874)  # get video_info_by_id 

    test_tag_id = video_info.tags[0]['id'] 
    api.video.tag.remove_from_video(test_tag_id)  # remove tag example
    api.video.tag.update_value(test_tag_id, 33221)  # update tag value
    api.video.tag.update_frame_range(test_tag_id, [0, 15])  # update tag frame range
"""



import asyncio
import copy
import re
import sys

from asgiref.sync import async_to_sync
from tqdm import tqdm

from supervisely.app import DataJson
from supervisely.app.widgets import Widget
from supervisely.sly_logger import logger, EventType


def extract_by_regexp(regexp, string):
    result = re.search(regexp, string)
    if result is not None:
        return result.group(0)
    else:
        return ''


class _slyProgressBarIO:
    def __init__(self, widget_id, message=None, total_provided=None):
        self.widget_id = widget_id

        self.progress = {
            'percent': 0,
            'info': '',
            'message': message,
            'status': None
        }

        self.prev_state = self.progress.copy()
        self.total_provided = total_provided

    def print_progress_to_supervisely_tasks_section(self):
        '''
        Logs a message with level INFO on logger. Message contain type of progress, subtask message, currtnt and total number of iterations
        '''
        extra = {
            'event_type': EventType.PROGRESS,
            'subtask': self.progress.get('message', None),
            'current': int(self.progress.get('percent', 0)),
            'total': 100,
        }

        gettrace = getattr(sys, "gettrace", None)
        in_debug_mode = gettrace is not None and gettrace()

        if not in_debug_mode and self.total_provided:
            logger.info('progress', extra=extra)

    def write(self, s):
        new_text = s.strip().replace('\r', '')
        if len(new_text) != 0:
            if self.total_provided:
                self.progress['percent'] = extract_by_regexp(r'\d*\%', new_text).replace('%', '')
                self.progress['info'] = extract_by_regexp(r'(\d+(?:\.\d+\w+)?)*/.*\]', new_text)
            else:
                self.progress['percent'] = 50
                self.progress['info'] = extract_by_regexp(r'(\d+(?:\.\d+\w+)?)*.*\]', new_text)

    def flush(self):
        if self.prev_state != self.progress:
            if self.progress['percent'] != '' and self.progress['info'] != '':
                self.print_progress_to_supervisely_tasks_section()

                for key, value in self.progress.items():
                    DataJson()[f'{self.widget_id}'][key] = value

                async_to_sync(DataJson().synchronize_changes)()

                self.prev_state = copy.deepcopy(self.progress)

    def __del__(self):
        DataJson()[f'{self.widget_id}']['status'] = "success"
        DataJson()[f'{self.widget_id}']['percent'] = 100
        self.progress['percent'] = 100

        self.print_progress_to_supervisely_tasks_section()

        async_to_sync(DataJson().synchronize_changes)()


class SlyTqdm(Widget):
    def __init__(self,
                 message: str = None,
                 show_percents: bool = False,
                 widget_id: str = None):
        """
        Wrapper for classic tqdm progress bar.

            Parameters
            ----------
            identifier  : int, required
                HTML element identifier
            message  : int, optional
                Text message which displayed in HTML


            desc, total, leave, ncols, ... :
                Like in tqdm

        """
        self.message = message
        self.show_percents = show_percents

        super().__init__(widget_id=widget_id, file_path=__file__)

    def __call__(self, iterable=None, message=None, desc=None, total=None, leave=True, ncols=None,
                 mininterval=1.0, maxinterval=10.0, miniters=None, ascii=False, disable=False, unit='it',
                 unit_scale=False, dynamic_ncols=False, smoothing=0.3, bar_format=None, initial=0, position=None,
                 postfix=None, unit_divisor=1000, gui=False, **kwargs):

        total_provided = True if (iterable is not None and hasattr(iterable, '__len__') and len(
            iterable) is not None) or total is not None else False

        sly_io = _slyProgressBarIO(self.widget_id, message, total_provided)

        return tqdm(iterable=iterable, desc=desc, total=total, leave=leave, file=sly_io, ncols=ncols,
                    mininterval=mininterval,
                    maxinterval=maxinterval, miniters=miniters, ascii=ascii, disable=disable, unit=unit,
                    unit_scale=unit_scale, dynamic_ncols=dynamic_ncols, smoothing=smoothing, bar_format=bar_format,
                    initial=initial, position=position, postfix=postfix, unit_divisor=unit_divisor,
                    gui=gui, **kwargs)

    def get_json_data(self):
        return {
            'percent': 0,
            'info': None,
            'message': self.message,
            'status': None,
            'show_percents': self.show_percents
        }

    def get_json_state(self):
        return None



from supervisely.app.widgets.widget import Widget
from supervisely.app.widgets.radio_table.radio_table import RadioTable
from supervisely.app.widgets.notification_box.notification_box import NotificationBox
from supervisely.app.widgets.done_label.done_label import DoneLabel
from supervisely.app.widgets.sly_tqdm.sly_tqdm import SlyTqdm

# from supervisely.app.widgets.restart_step.restart_step import RestartStep



# coding: utf-8

import os
from PIL import ImageFont
from supervisely.io.fs import get_file_ext, file_exists


FONT_EXTENSION = '.ttf'
DEFAULT_FONT_FILE_NAME = 'DejaVuSansMono.ttf'
_fonts = {}  # (font name, size) -> font


def _get_font_path_by_name(font_file_name: str) -> str:
    """
    Walk over systems fonts paths and match with given `font_file_name`.
    Args:
        font_file_name: font filename

    Returns:
        full path of requested by name font or None if file not found in system paths

    """
    import matplotlib.font_manager as fontman
    fonts_paths_list = fontman.findSystemFonts()
    for path in fonts_paths_list:
        if os.path.basename(path) == font_file_name:
            return path
    return None


def load_font(font_file_name: str, font_size: int = 12) -> ImageFont.FreeTypeFont:
    """
    Set global font true-type for drawing.
    Args:
        font_file_name: name of font file (example: 'DejaVuSansMono.ttf')
        font_size: selected font size
    Returns:
        loaded from file font
    """
    if get_file_ext(font_file_name) == FONT_EXTENSION:
        font_path = _get_font_path_by_name(font_file_name)
        if (font_path is not None) and file_exists(font_path):
            return ImageFont.truetype(font_path, font_size, encoding='utf-8')
        else:
            raise ValueError(
                'Font file "{}" not found in system paths. Try to set another font.'.format(font_file_name))
    else:
        raise ValueError('Supported only TrueType fonts!')


def get_font(font_file_name: str = None, font_size: int = 12) -> ImageFont.FreeTypeFont:
    """
    Args:
        font_file_name: name of font file (example: 'DejaVuSansMono.ttf')
        font_size: selected font size
    Returns:
        font for drawing
    """
    if font_file_name is None:
        font_file_name = DEFAULT_FONT_FILE_NAME

    font_key = (font_file_name, font_size)
    if font_key not in _fonts:
        _fonts[font_key] = load_font(font_file_name, font_size)
    return _fonts[font_key]


def get_readable_font_size(img_size) -> int:
    '''
    Get size of font for image with given sizes
    :param img_size: size of image
    :return: size of font
    '''
    minimal_font_size = 6
    base_font_size = 14
    base_image_size = 512
    return max(minimal_font_size, round(base_font_size * (img_size[0] + img_size[1]) // 2) // base_image_size)


# coding: utf-8

from supervisely.api.module_api import ApiField, ModuleApi


class PluginApi(ModuleApi):
    @staticmethod
    def info_sequence():
        return [ApiField.ID,
                ApiField.NAME,
                ApiField.DESCRIPTION,
                ApiField.TYPE,
                ApiField.DEFAULT_VERSION,
                ApiField.DOCKER_IMAGE,
                ApiField.README,
                ApiField.CONFIGS,
                ApiField.VERSIONS,
                ApiField.CREATED_AT,
                ApiField.UPDATED_AT]

    @staticmethod
    def info_tuple_name():
        return 'PluginInfo'

    def get_list(self, team_id, filters=None):
        '''
        :param team_id: int
        :param filters: list
        :return: list of plugins from the given team
        '''
        return self.get_list_all_pages('plugins.list',  {ApiField.TEAM_ID: team_id, ApiField.FILTER: filters or []})

    def get_info_by_id(self, team_id, plugin_id):
        '''
        :param team_id: int
        :param plugin_id: int
        :return: information about plugin from given plugin id in given team
        '''
        filters = [{"field": ApiField.ID, "operator": "=", "value": plugin_id}]
        return self._get_info_by_filters(team_id, filters)



# coding: utf-8

from supervisely.api.module_api import ApiField, ModuleNoParent, UpdateableModule
from typing import List
from supervisely.sly_logger import logger


#@TODO - umar will add meta with review status and duration
class ActivityAction:
    LOGIN = 'login'
    LOGOUT = 'logout'
    CREATE_PROJECT = 'create_project'
    UPDATE_PROJECT = 'update_project'
    DISABLE_PROJECT = 'disable_project'
    RESTORE_PROJECT = 'restore_project'
    CREATE_DATASET = 'create_dataset'
    UPDATE_DATASET = 'update_dataset'
    DISABLE_DATASET = 'disable_dataset'
    RESTORE_DATASET = 'restore_dataset'
    CREATE_IMAGE = 'create_image'
    UPDATE_IMAGE = 'update_image'
    DISABLE_IMAGE = 'disable_image'
    RESTORE_IMAGE = 'restore_image'
    CREATE_FIGURE = 'create_figure'
    UPDATE_FIGURE = 'update_figure'
    DISABLE_FIGURE = 'disable_figure'
    RESTORE_FIGURE = 'restore_figure'
    CREATE_CLASS = 'create_class'
    UPDATE_CLASS = 'update_class'
    DISABLE_CLASS = 'disable_class'
    RESTORE_CLASS = 'restore_class'
    CREATE_BACKUP = 'create_backup'
    EXPORT_PROJECT = 'export_project'
    MODEL_TRAIN = 'model_train'
    MODEL_INFERENCE = 'model_inference'
    CREATE_PLUGIN = 'create_plugin'
    DISABLE_PLUGIN = 'disable_plugin'
    RESTORE_PLUGIN = 'restore_plugin'
    CREATE_NODE = 'create_node'
    DISABLE_NODE = 'disable_node'
    RESTORE_NODE = 'restore_node'
    CREATE_WORKSPACE = 'create_workspace'
    DISABLE_WORKSPACE = 'disable_workspace'
    RESTORE_WORKSPACE = 'restore_workspace'
    CREATE_MODEL = 'create_model'
    DISABLE_MODEL = 'disable_model'
    RESTORE_MODEL = 'restore_model'
    ADD_MEMBER = 'add_member'
    REMOVE_MEMBER = 'remove_member'
    LOGIN_TO_TEAM = 'login_to_team'
    ATTACH_TAG = 'attach_tag'
    UPDATE_TAG_VALUE = 'update_tag_value'
    DETACH_TAG = 'detach_tag'
    ANNOTATION_DURATION = 'annotation_duration'
    IMAGE_REVIEW_STATUS_UPDATED = 'image_review_status_updated'

    # case #1 - labeler pressed "finish image" button in labeling job
    # action: IMAGE_REVIEW_STATUS_UPDATED -> meta["reviewStatus"] == 'done'

    # case #2 - reviewer pressed "accept" or "reject" button
    # action: IMAGE_REVIEW_STATUS_UPDATED -> meta["reviewStatus"] == 'accepted' or 'rejected'

    # possible review statuses:
    # 'done' - i.e. labeler finished the image,
    # 'accepted' - reviewer
    # 'rejected' - reviewer

    # case #3 duration
    # action: ANNOTATION_DURATION -> meta["duration"] e.g. meta-> {"duration": 30} in seconds


class TeamApi(ModuleNoParent, UpdateableModule):
    @staticmethod
    def info_sequence():
        return [ApiField.ID,
                ApiField.NAME,
                ApiField.DESCRIPTION,
                ApiField.ROLE,
                ApiField.CREATED_AT,
                ApiField.UPDATED_AT]

    @staticmethod
    def info_tuple_name():
        return 'TeamInfo'

    def __init__(self, api):
        ModuleNoParent.__init__(self, api)
        UpdateableModule.__init__(self, api)

    def get_list(self, filters=None):
        '''
        :param filters: list
        :return: list of all teams
        '''
        return self.get_list_all_pages('teams.list',  {ApiField.FILTER: filters or []})

    def get_info_by_id(self, id):
        '''
        :param id: int
        :return: team metainformation with given id
        '''
        return self._get_info_by_id(id, 'teams.info')

    def create(self, name, description="", change_name_if_conflict=False):
        '''
        Create team with given name
        :param name: str
        :param description: str
        :param change_name_if_conflict: bool
        :return: team metainformation
        '''
        effective_name = self._get_effective_new_name(name=name, change_name_if_conflict=change_name_if_conflict)
        response = self._api.post('teams.add', {ApiField.NAME: effective_name, ApiField.DESCRIPTION: description})
        return self._convert_json_info(response.json())

    def _get_update_method(self):
        return 'teams.editInfo'

    def get_activity(self, team_id,
                     filter_user_id=None, filter_project_id=None, filter_job_id=None, filter_actions=None,
                     progress_cb=None, start_date=None, end_date=None):

        from datetime import datetime, timedelta

        filters = []
        if filter_user_id is not None:
            filters.append({"field": ApiField.USER_ID, "operator": "=", "value": filter_user_id})
        if filter_project_id is not None:
            filters.append({"field": ApiField.PROJECT_ID, "operator": "=", "value": filter_project_id})
        if filter_job_id is not None:
            filters.append({"field": ApiField.JOB_ID, "operator": "=", "value": filter_job_id})
        if filter_actions is not None:
            if type(filter_actions) is not list:
                raise TypeError(
                    "type(filter_actions) is {!r}. But has to be of type {!r}".format(type(filter_actions), list))
            filters.append({"field": ApiField.TYPE, "operator": "in", "value": filter_actions})

        def _add_dt_filter(filters, dt, op):
            dt_iso = None
            if dt is None:
                return
            if type(dt) is str:
                dt_iso = dt
            elif type(dt) is datetime:
                dt_iso = dt.isoformat()
            else:
                raise TypeError('DT type must be string in ISO8601 format or datetime, not {}'.format(type(dt)))
            filters.append({"field": ApiField.DATE, "operator": op, "value": dt_iso})

        _add_dt_filter(filters, start_date, ">=")
        _add_dt_filter(filters, end_date, "<=")

        method = 'teams.activity'
        data = {ApiField.TEAM_ID: team_id, ApiField.FILTER: filters}
        first_response = self._api.post(method, data)
        first_response = first_response.json()

        total = first_response['total']
        per_page = first_response['perPage']
        pages_count = first_response['pagesCount']
        results = first_response['entities']

        if progress_cb is not None:
            progress_cb(len(results), total)
        if pages_count == 1 and len(first_response['entities']) == total:
            pass
        else:
            for page_idx in range(2, pages_count + 1):
                temp_resp = self._api.post(method, {**data, 'page': page_idx, 'per_page': per_page})
                temp_items = temp_resp.json()['entities']
                results.extend(temp_items)
                if progress_cb is not None:
                    progress_cb(len(results), total)
            if len(results) != total:
                logger.warn(f"Method '{method}': new events were created during pagination, "
                            f"downloaded={len(results)}, total={total}")

        return results



# coding: utf-8

import collections
import operator
import numpy as np
from typing import List, Callable

from supervisely.annotation.annotation import Annotation
from supervisely.annotation.label import Label
from supervisely.annotation.obj_class import ObjClass
from supervisely.geometry.bitmap import Bitmap, SkeletonizeMethod
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.polyline import Polyline
from supervisely.geometry.rectangle import Rectangle
from supervisely.geometry.point_location import PointLocation


def skeletonize_bitmap(ann: Annotation, classes: List[str], method_id: SkeletonizeMethod) -> Annotation:
    """
    Extracts skeletons from bitmap figures.

    Args:
        ann: Input annotation.
        classes: List of classes to skeletonize.
        method_id: Algorithm of processing. See supervisely.geometry.bitmap.SkeletonizeMethod enum.
    Returns:
        Annotation with skeletonized labels.
    """
    def _skel(label: Label):
        if label.obj_class.name not in classes:
            return [label]

        if not isinstance(label.geometry, Bitmap):
            raise RuntimeError('Input class must be a Bitmap.')

        return [label.clone(geometry=label.geometry.skeletonize(method_id))]

    return ann.transform_labels(_skel)


def approximate_vector(ann: Annotation, classes: List[str], epsilon: float) -> Annotation:
    """
    Approximates vector figures: lines and polygons.

    Args:
        ann: Input annotations.
        classes: List of classes to apply transformation.
        epsilon: Approximation accuracy (maximum distance between the original curve and its approximation)
    Returns:
        Annotation with approximated vector figures of selected classes.
    """
    def _approx(label: Label):
        if label.obj_class.name not in classes:
            return [label]

        if not isinstance(label.geometry, (Polygon, Polyline)):
            raise RuntimeError('Input class must be a Polygon or a Line.')

        return [label.clone(geometry=label.geometry.approx_dp(epsilon))]

    return ann.transform_labels(_approx)


def add_background(ann: Annotation, bg_class: ObjClass) -> Annotation:
    """
    Adds background rectangle (size equals to image size) to annotation.

    Args:
        ann: Input annotation.
        bg_class: ObjClass instance for background class label.
    Returns:
        Annotation with added background rectangle.
    """
    img_size = ann.img_size
    rect = Rectangle(0, 0, img_size[0] - 1, img_size[1] - 1)
    new_label = Label(rect, bg_class)
    return ann.add_label(new_label)


def drop_object_by_class(ann: Annotation, classes: List[str]) -> Annotation:
    """
    Removes labels of specified classes from annotation.

    Args:
        ann: Input annotation.
        classes: List of classes to remove.
    Returns:
        Annotation with removed labels of specified classes.
    """
    def _filter(label: Label):
        if label.obj_class.name in classes:
            return [label]
        return []
    return ann.transform_labels(_filter)


def filter_objects_by_area(ann: Annotation, classes: List[str], comparator=operator.lt,
                           thresh_percent: float = None) -> Annotation:  # @ TODO: add size mode
    """
    Deletes labels less (or greater) than specified percentage of image area.

    Args
        ann: Input annotation.
        classes: List of classes to filter.
        comparator: Comparison function.
        thresh_percent: Threshold percent value of image area.
    Returns:
        Annotation containing filtered labels.
    """
    imsize = ann.img_size
    img_area = float(imsize[0] * imsize[1])

    def _del_filter_percent(label: Label):
        if label.obj_class.name in classes:
            fig_area = label.area
            area_percent = 100.0 * fig_area / img_area
            if comparator(area_percent, thresh_percent):  # satisfied condition
                return []  # action 'delete'
        return [label]

    return ann.transform_labels(imsize, _del_filter_percent)


def bitwise_mask(ann: Annotation, class_mask: str, classes_to_correct: List[str],
                 bitwise_op: Callable[[np.ndarray, np.ndarray], np.ndarray] = np.logical_and) -> Annotation:
    """
    Performs bitwise operation between two masks. Uses one target mask to correct all others.

    Args
        ann: Input annotation.
        class_mask: Class name of target mask.
        classes_to_correct: List of classes which will be corrected using target mask.
        bitwise_op: Bitwise numpy function to process masks.For example: "np.logical_or", "np.logical_and",
         "np.logical_xor".
    Returns:
        Annotation containing corrected Bitmaps.
    """
    imsize = ann.img_size

    def find_mask_class(labels, class_mask_name):
        for label in labels:
            if label.obj_class.name == class_mask_name:
                if not isinstance(label.geometry, Bitmap):
                    raise RuntimeError('Class <{}> must be a Bitmap.'.format(class_mask_name))
                return label

    mask_label = find_mask_class(ann.labels, class_mask)
    if mask_label is not None:
        target_original, target_mask = mask_label.geometry.origin, mask_label.geometry.data
        full_target_mask = np.full(imsize, False, bool)

        full_target_mask[target_original.row:target_original.row + target_mask.shape[0],
                         target_original.col:target_original.col + target_mask.shape[1]] = target_mask

        def perform_op(label):
            if label.obj_class.name not in classes_to_correct or label.obj_class.name == class_mask:
                return [label]

            if not isinstance(label.geometry, Bitmap):
                raise RuntimeError('Input class must be a Bitmap.')

            new_geom = label.geometry.bitwise_mask(full_target_mask, bitwise_op)
            return [label.clone(geometry=new_geom)] if new_geom is not None else []

        res_ann = ann.transform_labels(perform_op)
    else:
        res_ann = ann.clone()

    return res_ann


def find_contours(ann: Annotation, classes_mapping: dict) -> Annotation:  # @TODO: approximation dropped
    """

    Args:
        ann: Input annotation.
        classes_mapping: Dict matching source class names and new ObjClasses
    Returns:
        Annotation with Bitmaps converted to contours Polygons.
    """
    def to_contours(label: Label):
        new_obj_cls = classes_mapping.get(label.obj_class.name)
        if new_obj_cls is None:
            return [label]
        if not isinstance(label.geometry, Bitmap):
            raise RuntimeError('Input class must be a Bitmap.')

        return [Label(geometry=geom, obj_class=new_obj_cls) for geom in label.geometry.to_contours()]

    return ann.transform_labels(to_contours)


def extract_labels_from_mask(mask: np.ndarray, color_id_to_obj_class: collections.Mapping) -> list:
    """
    Extract multiclass instances from grayscale mask and save it to labels list.
    Args:
        mask: multiclass grayscale mask
        color_id_to_obj_class: dict of objects classes assigned to color id (e.g. {1: ObjClass('cat), ...})
    Returns:
        list of labels with bitmap geometry
    """
    from skimage import measure
    from scipy import ndimage
    
    zero_offset = 1 if 0 in color_id_to_obj_class else 0
    if zero_offset > 0:
        mask = mask + zero_offset

    labeled, labels_count = measure.label(mask, connectivity=1, return_num=True)
    objects_slices = ndimage.find_objects(labeled)
    labels = []

    for object_index, slices in enumerate(objects_slices, start=1):
        crop = mask[slices]
        sub_mask = crop * (labeled[slices] == object_index).astype(np.int)

        class_index = np.max(sub_mask) - zero_offset

        if class_index in color_id_to_obj_class:
            bitmap = Bitmap(data=sub_mask.astype(np.bool), origin=PointLocation(slices[0].start, slices[1].start))
            label = Label(geometry=bitmap, obj_class=color_id_to_obj_class.get(class_index))
            labels.append(label)
    return labels



# coding: utf-8

import numpy as np

from supervisely.geometry.point_location import PointLocation
from supervisely.geometry.rectangle import Rectangle
from supervisely.geometry.bitmap import Bitmap
from supervisely.annotation.json_geometries_map import GET_GEOMETRY_FROM_STR
from supervisely.geometry.polyline import Polyline
from supervisely.geometry.polygon import Polygon


def geometry_to_bitmap(geometry, radius: int = 0, crop_image_shape: tuple = None) -> list:
    """
    Args:
        geometry: Geometry type which implemented 'draw', 'translate' and 'to_bbox` methods
        radius: half of thickness of drawed vector elements
        crop_image_shape: if not None - crop bitmap object by this shape (HxW)
    Returns:
        Bitmap (geometry) object
    """

    thickness = radius + 1

    bbox = geometry.to_bbox()
    extended_bbox = Rectangle(top=bbox.top - radius,
                              left=bbox.left - radius,
                              bottom=bbox.bottom + radius,
                              right=bbox.right + radius)
    bitmap_data = np.full(shape=(extended_bbox.height, extended_bbox.width), fill_value=False)
    geometry = geometry.translate(-extended_bbox.top, -extended_bbox.left)
    geometry.draw(bitmap_data, color=True, thickness=thickness)

    origin = PointLocation(extended_bbox.top, extended_bbox.left)
    bitmap_geometry = Bitmap(data=bitmap_data, origin=origin)
    if crop_image_shape is not None:
        crop_rect = Rectangle.from_size(*crop_image_shape)
        return bitmap_geometry.crop(crop_rect)
    return [bitmap_geometry]


def get_effective_nonoverlapping_masks(geometries, img_size=None):
    '''
    Find nonoverlapping objects from given list of geometries
    :param geometries: list of geometry type objects(Point, Polygon, PolyLine, Bitmap etc.)
    :param img_size: tuple or list of integers
    :return: list of bitmaps, numpy array
    '''
    if img_size is None:
        if len(geometries) > 0:
            common_bbox = Rectangle.from_geometries_list(geometries)
            img_size = (common_bbox.bottom + 1, common_bbox.right + 1)
        else:
            img_size = (0,0)
    canvas = np.full(shape=img_size, fill_value=len(geometries), dtype=np.int32)

    for idx, geometry in enumerate(geometries):
        geometry.draw(canvas, color=idx)
    result_masks = []
    for idx, geometry in enumerate(geometries):
        effective_indicator = (canvas == idx)
        if np.any(effective_indicator):
            result_masks.append(Bitmap(effective_indicator))
        else:
            result_masks.append(None)
    return result_masks, canvas


def deserialize_geometry(geometry_type_str, geometry_json):
    '''
    Get geometry from json format
    :param geometry_type_str: str
    :param geometry_json: geometry in json format
    :return: geometry type object(Point, Polygon, PolyLine, Bitmap etc.)
    '''
    geometry_type = GET_GEOMETRY_FROM_STR(geometry_type_str)
    geometry = geometry_type.from_json(geometry_json)
    return geometry


def geometry_to_polygon(geometry, approx_epsilon=None):
    if type(geometry) not in (Rectangle, Polyline, Polygon, Bitmap):
        raise KeyError('Can not convert {} to {}'.format(geometry.geometry_name(), Polygon.__name__))

    if type(geometry) == Rectangle:
        return [Polygon(geometry.corners, [])]

    if type(geometry) == Polyline:
        return [Polygon(geometry.exterior, [])]

    if type(geometry) == Polygon:
        return [geometry]

    if type(geometry) == Bitmap:
        new_geometries = geometry.to_contours()
        if approx_epsilon is None:
            approx_epsilon = 1

        new_geometries = [g.approx_dp(approx_epsilon) for g in new_geometries]
        return new_geometries


# coding: utf-8

from supervisely.api.module_api import ApiField, ModuleApi, UpdateableModule


class WorkspaceApi(ModuleApi, UpdateableModule):
    @staticmethod
    def info_sequence():
        return [ApiField.ID,
                ApiField.NAME,
                ApiField.DESCRIPTION,
                ApiField.TEAM_ID,
                ApiField.CREATED_AT,
                ApiField.UPDATED_AT]

    @staticmethod
    def info_tuple_name():
        return 'WorkspaceInfo'

    def __init__(self, api):
        ModuleApi.__init__(self, api)
        UpdateableModule.__init__(self, api)

    def get_list(self, team_id, filters=None):
        '''
        :param team_id: int
        :param filters: list
        :return: list of all the workspaces in the selected team
        '''
        return self.get_list_all_pages('workspaces.list',  {ApiField.TEAM_ID: team_id, ApiField.FILTER: filters or []})

    def get_info_by_id(self, id):
        '''
        :param id: int
        :return: workspace metadata
        '''
        return self._get_info_by_id(id, 'workspaces.info')

    def create(self, team_id, name, description="", change_name_if_conflict=False):
        '''
        Create workspaces in team with given id
        :param team_id: int
        :param name: str
        :param description: str
        :param change_name_if_conflict: bool
        :return: new workspace metadata
        '''
        effective_name = self._get_effective_new_name(
            parent_id=team_id, name=name, change_name_if_conflict=change_name_if_conflict)
        response = self._api.post('workspaces.add', {ApiField.TEAM_ID: team_id,
                                                     ApiField.NAME: effective_name,
                                                     ApiField.DESCRIPTION: description})
        return self._convert_json_info(response.json())

    def _get_update_method(self):
        return 'workspaces.editInfo'



import inspect
from collections import OrderedDict
from supervisely.sly_logger import logger
from supervisely.annotation.annotation import Annotation
from supervisely.project.project_meta import ProjectMeta
import numpy as np


def create_aug_info(category_name, aug_name, params, sometimes: float = None):
    clean_params = params
    #clean_params = remove_unexpected_arguments(category_name, aug_name, params)
    res = {
        "category": category_name,
        "name": aug_name,
        "params": clean_params,
    }
    if sometimes is not None:
        if type(sometimes) is not float or not (0.0 <= sometimes <= 1.0):
            raise ValueError(f"sometimes={sometimes}, type != {type(sometimes)}")
        res["sometimes"] = sometimes
    res["python"] = aug_to_python(res)
    return res


def aug_to_python(aug_info):
    pstr = ""
    for name, value in aug_info["params"].items():
        v = value
        if type(v) is list:  #name != 'nb_iterations' and
            v = (v[0], v[1])
        if type(value) is str:
            pstr += f"{name}='{v}', "
        else:
            pstr += f"{name}={v}, "
    method_py = f"iaa.{aug_info['category']}.{aug_info['name']}({pstr[:-2]})"

    res = method_py
    if "sometimes" in aug_info:
        res = f"iaa.Sometimes({aug_info['sometimes']}, {method_py})"
    return res


def pipeline_to_python(aug_infos, random_order=False):
    template = \
"""import imgaug.augmenters as iaa

seq = iaa.Sequential([
{}
], random_order={})
"""
    py_lines = []
    for info in aug_infos:
        line = aug_to_python(info)
        _validate = info["python"]
        if line != _validate:
            raise ValueError("Generated python line differs from the one from config: \n\n{!r}\n\n{!r}"
                             .format(line, _validate))
        py_lines.append(line)
    res = template.format('\t' + ',\n\t'.join(py_lines), random_order)
    return res


def get_default_params_by_function(f):
    params = []
    method_info = inspect.signature(f)
    for param in method_info.parameters.values():
        formatted = str(param)
        if 'deprecated' in formatted or 'seed=None' in formatted or 'name=None' in formatted:
            continue
        if param.default == inspect._empty:
            continue
        params.append({
            "pname": param.name,
            "default": param.default
        })
    return params


def get_default_params_by_name(category_name, aug_name):
    func = get_function(category_name, aug_name)
    defaults = get_default_params_by_function(func)
    return defaults


def get_function(category_name, aug_name):
    import imgaug.augmenters as iaa
    try:
        submodule = getattr(iaa, category_name)
        aug_f = getattr(submodule, aug_name)
        return aug_f
    except Exception as e:
        logger.error(repr(e))
        # raise e
        return None


def build_pipeline(aug_infos, random_order=False):
    import imgaug.augmenters as iaa
    pipeline = []
    for aug_info in aug_infos:
        category_name = aug_info["category"]
        aug_name = aug_info["name"]
        params = aug_info["params"]

        aug_func = get_function(category_name, aug_name)
        # TODO: hotfix:
        if aug_name == "CropAndPad":
            params["percent"] = tuple(params["percent"])

        aug = aug_func(**params)

        sometimes = aug_info.get("sometimes", None)
        if sometimes is not None:
            aug = iaa.meta.Sometimes(sometimes, aug)
        pipeline.append(aug)
    augs = iaa.Sequential(pipeline, random_order=random_order)
    return augs


def build(aug_info):
    return build_pipeline([aug_info])


def remove_unexpected_arguments(category_name, aug_name, params):
    # to avoid this:
    # TypeError: f() got an unexpected keyword argument 'b'
    defaults = get_default_params_by_name(category_name, aug_name)
    allowed_names = [d["pname"] for d in defaults]

    res = OrderedDict()
    for name, value in params.items():
        if name in allowed_names:
            res[name] = value
    return res


def _apply(augs, img, boxes=None, masks=None):
    import imgaug.augmenters as iaa
    augs: iaa.Sequential
    res = augs(images=[img], bounding_boxes=boxes, segmentation_maps=masks)
    #return image, boxes, masks
    return res[0][0], res[1], res[2]


def apply(augs, meta: ProjectMeta, img, ann: Annotation, segmentation_type='semantic'):
    # @TODO: save object tags

    # works for rectangles
    det_meta, det_mapping = meta.to_detection_task(convert_classes=False)
    det_ann = ann.to_detection_task(det_mapping)
    ia_boxes = det_ann.bboxes_to_imgaug()

    # works for polygons and bitmaps
    seg_meta, seg_mapping = meta.to_segmentation_task()
    seg_ann = ann.to_nonoverlapping_masks(seg_mapping)

    if segmentation_type == 'semantic':
        seg_ann = seg_ann.to_segmentation_task()
        class_to_index = {obj_class.name: idx for idx, obj_class in enumerate(seg_meta.obj_classes, start=1)}
        index_to_class = {v: k for k, v in class_to_index.items()}
    elif segmentation_type == 'instance':
        class_to_index = None
        index_to_class = {idx: label.obj_class.name for idx, label in enumerate(seg_ann.labels, start=1)}
    elif segmentation_type == 'panoptic':
        raise NotImplementedError

    ia_masks = seg_ann.masks_to_imgaug(class_to_index)

    res_meta = det_meta.merge(seg_meta)  # TagMetas should be preserved

    res_img, res_ia_boxes, res_ia_masks = _apply(augs, img, ia_boxes, ia_masks)
    res_ann = Annotation.from_imgaug(res_img,
                                     ia_boxes=res_ia_boxes, ia_masks=res_ia_masks,
                                     index_to_class=index_to_class, meta=res_meta)
    # add image tags
    res_ann = res_ann.clone(img_tags=ann.img_tags)
    return res_meta, res_img, res_ann


def apply_to_image(augs, img):
    res_img, _, _ = _apply(augs, img, None, None)
    return res_img


def apply_to_image_and_mask(augs, img, mask):
    from imgaug.augmentables.segmaps import SegmentationMapsOnImage
    segmaps = SegmentationMapsOnImage(mask, shape=img.shape[:2])
    res_img, _, res_segmaps = _apply(augs, img, masks=segmaps)
    res_mask = res_segmaps.get_arr()
    if res_img.shape[:2] != res_mask.shape[:2]:
        raise ValueError(f"Image and mask have different shapes "
                         f"({res_img.shape[:2]} != {res_mask.shape[:2]}) after augmentations. "
                         f"Please, contact tech support")
    return res_img, res_mask




# coding: utf-8
from supervisely.api.module_api import ApiField
from supervisely.api.pointcloud.pointcloud_api import PointcloudApi
from supervisely.api.pointcloud.pointcloud_episode_annotation_api import PointcloudEpisodeAnnotationAPI


class PointcloudEpisodeApi(PointcloudApi):
    def __init__(self, api):
        super().__init__(api)
        self.annotation = PointcloudEpisodeAnnotationAPI(api)

    def _convert_json_info(self, info: dict, skip_missing=True):
        res = super()._convert_json_info(info, skip_missing=skip_missing)
        if res.meta is not None:
            return res._replace(frame=res.meta[ApiField.FRAME])
        else:
            raise RuntimeError('Error with point cloud meta or API version. Please, contact support')

    def get_frame_name_map(self, dataset_id):
        pointclouds = self.get_list(dataset_id)
        return {x.frame: x.name for x in pointclouds}


